import time

def time_to_date(timestamp):
    """
    timestamp to date.
    :param timestamp :int,e.g.1537535021
    :return:Year-Month-Day Hour:Minute:Second
    """
    timearr = time.localtime(timestamp)
    otherStyleTime = time.strftime("%Y-%m-%d %H:%M:%S", timearr)
    return  otherStyleTime


#coding:utf-8
import copy
import requests
import asyncio
from contextlib import closing

async def send_async_http(session,method,url,*,
                             retries=1,
                             interval=1,
                             wait_factor=2,
                             timeout=0,
                             path = None,
                             success_callback=None,
                             fail_callback=None,
                             **kwargs) -> dict:
    """
    Send an async http request and implement the retry mechanism.
    Once the requesting operation failed,it will delay a interval time
    for the next retry.
    :param session:asynchronous request session
    :param method:request method
    :param url:request url
    :param retries:how many it will retry
    :param interval:a delay interval secs.
    :param wait_factor:wait factor,every retry failed will multiply by it to delay for
    next retry,recommended 1<wf<2
    :param timeout:requesting timeout
    :param success_callback:callback for successes
    :param fail_callback:callback for fails
    :param kwargs:other kwargs
    :param path:file save path
    :return:result
    """
    ret = {'resp':None,'body':None,'code':-1,
           'exception':None,'tries':-1}
    wait_interval = interval
    if method.lower() not in ['get','head','post','put']:
        return ret
    if retries == -1:  # -1 means retry unlimited times
        attempt = -1
    elif retries == 0:  # 0 means no retry
        attempt = 1
    else:
        attempt = retries + 1
    while attempt != 0:
        size = 0
        try:
            if path:
                loop = asyncio.get_running_loop()
                try:
                    resp = await loop.run_in_executor(None,download,url,path,timeout,kwargs)
                    size = resp
                    body = True
                    code = 200
                except requests.exceptions.Timeout:
                    raise TimeoutError
            else:
                async with getattr(session,method.lower())(url,timeout=timeout,**kwargs) as response:
                    code = response.status
                    resp = response
                    body = await response.read()
            ret.update({'resp': resp, 'body':body,'code': code,
                        'tries': retries - attempt+1,'size':size})
            if success_callback:
                success_callback(ret)
            return ret
        except Exception as e:
            ret['exception'] = e.__class__()
            ret['tries'] += 1
            await asyncio.sleep(wait_interval)
            wait_interval = wait_interval * wait_factor
        attempt-=1
    if fail_callback:
        fail_callback(ret)
    return ret

def download(url,filepath,timeout,kwargs):
    if timeout==False:
        timeout=None
    _kw = copy.deepcopy(kwargs)
    buffer = _kw.pop('buffer')
    with closing(requests.get(url, stream=True,timeout=timeout,**_kw)) as response:
        chunk_size = buffer
        data_count = 0
        with open(filepath,'wb') as f:
            for data in response.iter_content(chunk_size=chunk_size):
                f.write(data)
                data_count = data_count + len(data)
            return data_count



import asyncio
import time
from threading import Thread
from amipy.BaseClass import SpiderClientCommand
from amipy.cmd import _iter_specify_classes

class SpiderServer(Thread):

    def __init__(self,settings,spiders):
        super(SpiderServer,self).__init__()
        self.spiders = spiders
        self.settings = settings
        self.loop = asyncio.new_event_loop()
        self.host = settings['project'].SPIDER_SERVER_HOST
        self.port = settings['project'].SPIDER_SERVER_PORT
        self.tool_module = settings['project'].SPIDER_SERVER_COMMANDS_MODULE
        self.setDaemon(True)
        self.prompt = b"""* Spider-Client commands tool:
                    \r\n* use "help" to see all the commands usage."""

    async def _handle_request(self,reader:asyncio.StreamReader,writer:asyncio.StreamWriter):
        writer.write(self.prompt)
        while 1:
            writer.write(b'\r\n$amipy> ')
            client = writer.get_extra_info('peername')
            _c = ':'.join(str(i) for i in client)
            try:
                await writer.drain()
                data = await reader.readline()
                msg = data.decode().strip()
                if msg == 'quit':
                    print(f'*[Server] {time.ctime()} Connection closed at {_c}')
                    writer.close()
                    return
                elif msg:
                    resp = self.parse_opt(msg)
                    print(f'*[Server] {time.ctime()} Received "{msg}" from {_c}.')
                    writer.write(resp.encode('latin-1'))
            except Exception as e:
                print(f'*[Server] {time.ctime()} {e} at {_c}')
                writer.close()
            if not writer.is_closing():
                await writer.drain()
            else:
                writer.close()
                return

    def _pop_cmdname(self,msg):
        args = [i for i in msg.strip().split(' ') if i]
        import string
        for _,v in enumerate(args):
            if v and v[0] not in string.punctuation:
                args.pop(_)
                if  v in ['help','list'] or(args and args[0][0] not in string.punctuation):
                    return  v,args
        return None,None

    def _get_all_cmds(self,module):
        cmds = {}
        for cmd in _iter_specify_classes(module, cmdcls=SpiderClientCommand):
            cmdname = cmd.__module__.split('.')[-1]
            cmds[cmdname] = cmd
        return cmds

    def parse_opt(self,msg):
        cmdname,args = self._pop_cmdname(msg)
        cmds = self._get_all_cmds(self.tool_module)
        if not cmds.get(cmdname):
            return """\r\n* Command Usage:
            \r\n <option> [spider name] 
            \r\n or:  show spiders
            """
        data = cmds[cmdname]().parse(cmdname,args,self.spiders)
        return data

    def serve(self):
        coro = asyncio.start_server(self._handle_request,self.host,self.port,loop=self.loop)
        server = self.loop.run_until_complete(coro)
        addr,port = server.sockets[0].getsockname()
        print(f'* Spider server serving on {addr}:{port}.')
        print('* Press Ctrl+C to stop the crawling.\n')
        try:
            self.loop.run_forever()
        except (KeyboardInterrupt,StopAsyncIteration):
            print('Shutting down spider server.')
        server.close()
        self.loop.run_until_complete(server.wait_closed())
        self.loop.close()

    def run(self):
        asyncio.set_event_loop(self.loop)
        self.serve()







#coding:utf-8
import amipy
import json
import aiohttp
import w3lib.url as urltool
from amipy.exceptions import NotValidJsonContent

class mdict(dict):

   def __getattr__(self, item):
       try:
           a = self[item]
       except:
           return None
       return a

class Response(object):
    def __init__(self,url,
                 status=None,
                 headers=None,
                 request=None,
                 priority=0,
                 encoding='utf-8',
                 body=None,
                 exc = None,
                 cookies=None,
                 _resp=None
                 ):

        assert isinstance(_resp,aiohttp.ClientResponse) or _resp is None,\
            f'_resp of a Response must be a aiohttp.ClientResponse,' \
            f'got {type(_resp).__name__}.'

        assert isinstance(request, amipy.Request),\
            'not a valid Request for Response,got "%s".'%type(request).__name__

        if exc is not None:
            if not isinstance(exc, Exception):
                raise TypeError('Not an valid Exception for Response,got "%s".'
                                % type(exc).__name__)

        self.request = request
        self.spider = request.spider
        self.callback = request.callback
        self.errback = request.errback
        self.excback = request.excback
        self.priority = priority
        self._body = body
        self.exception = exc
        self.resp = _resp

        self.msg = _resp.reason if _resp else None
        self.headers = _resp.headers if _resp and headers is None else headers
        self.content_type = _resp.content_type if _resp else None
        self.history = _resp.history if _resp else None
        self.encoding = _resp.charset  if _resp else encoding
        self.__cookies = _resp.cookies if _resp else None
        self.status = _resp.status if  _resp and status is None else status
        self.http_ver = _resp.version if _resp else None
        self.request_info = _resp.request_info if _resp else None
        self.cookies = cookies if cookies else self.__cookies

        fingerprint = request.fingerprint
        self.resp_filter = bool(fingerprint) if  fingerprint != None \
            else self.spider.settings.BLOOMFILTER_HTML_ON
        self.fingerprint  = fingerprint if fingerprint != None and \
            not isinstance(fingerprint,bool) \
            else self.spider.fingerprint
        self.meta =mdict(request.kwargs_cb)
        self._set_url(url)

    def _set_url(self,url):
        self.url = urltool.canonicalize_url(
            urltool.safe_download_url(url),encoding=self.encoding)

    def text(self,encoding=None):
        encoding = encoding if encoding else self.encoding
        if encoding is None:
            encoding='utf-8'
        if isinstance(self._body,bytes):
            return str(self._body,encoding=encoding)
        return self._body

    def json(self):
        try:
            res  = json.loads(self._body)
            return res
        except json.decoder.JSONDecodeError as e:
            raise NotValidJsonContent(e)

    def read(self,encoding=None):
        encoding = encoding if encoding else self.encoding
        if encoding is None:
            encoding='utf-8'
        if isinstance(self._body,str):
            return bytes(self._body,encoding=encoding)
        return self._body

    @property
    def url(self):
        return self.__url

    @url.setter
    def url(self,url):
        #only starts with schema:file,http,https allowed to be a valid url
        if not urltool.is_url(url):
            raise ValueError('Not a valid url for Request.')
        else:
            self.__url = urltool.safe_download_url(url)


    def __str__(self):
        return '<Response obj at %s [status=%d url=%s] >'\
               %(hex(id(self)),self.status,self.url)




import re,os,string
import amipy
from amipy.BaseClass import Command
from amipy.core.loader import SpiderLoader
from amipy.util.file import copy_files
class AnyNameYouWant(Command):

    requires_project = True

    def handle(self,settings,opts, args):
        print(f'\n* Active Amipy project:{settings["project"].PROJECT_NAME}')
        print(f'* Creating a Spider.')
        spiders  = SpiderLoader(settings).load_all_spiders()
        _exists = []
        for i in spiders:
            i.close(save=False)
            _exists.append(i.name)
        name = re.sub('[%s ]'%string.punctuation,'',args[0])
        if name[0] not in string.ascii_letters:
            print(f'* Invalid name "{name}",please choose another name starts with letters only.')
            return
        if name in _exists:
            print(f'* Name "{name}" has already existed.Please choose another name for your Spider.')
            return
        else:
            _base = os.path.join(settings["project"].PATH,
                                 settings['project'].SPIDER_MODULE.replace('.',os.sep))
            destination = os.path.join(_base,name.title())
            src = os.path.join(amipy.__path__[0], 'templates.spider'.replace('.', os.sep))
            render_vars = {'spider_name': name,
                           'spider_path': destination,
                           'SpiderName':name.title()}
            render_files = {
                'settings.py.tpl': render_vars,
                'spider.py.tpl':render_vars
            }
            copy_files(src, destination, ('*.pyc', '*.pyd'), render=render_files)
            print(f'* Created a spider:{name}')
            print(f'* What to do next?')
            print(f'1. Now you can get inside the "spiders" folder.')
            print(f'2. And you will  see there\'s a "{name.title()}" spider folder.')
            print(f'3. Edit the spider.py under the spider folder.')
            print(f'4. Run the Spider.')

    @classmethod
    def short_desc(self):
        return 'create a new crawling spider.'

    def syntax(self):
        return ' <spider name> [options] <args> '




import amipy
from amipy.middlewares import Middleware
from amipy.util.proxy import gen_proxy,\
    is_proxy_valid,extract_ip_port
from amipy.log import getLogger
from w3lib.url import parse_url,is_url

class HttpProxyMiddleware(Middleware):

    inited = False
    invalid_pool = {}
    proxy_pool = set()
    logger = getLogger(__name__)

    def _proxy_invalid(self,proxy,url):
        domain = parse_url(url).netloc
        if proxy in self.invalid_pool:
            if domain in self.invalid_pool[proxy]:
                return True
            return False

    def process_request(self,request):
        if not request.spider.settings.HTTP_PROXY_ENABLE:
            request.proxy = None
            return request
        _type = request.down_type
        proxy = request.proxy
        url = request.url
        if proxy:
            if not is_proxy_valid(proxy):
                if request.spider.settings.HTTP_PROXY_FILL_ENABLE:
                    request.proxy = self.get_proxy(request)
                    if request.proxy:
                        self.logger.warn(f'Filling a new proxy {request.proxy} to {url}.')
                else:
                    self.logger.error(f'Not a valid http proxy:{proxy}')
                    request.proxy = None
                return request
            elif self._proxy_invalid(proxy,url):
                    self.logger.warn(f'Proxy {proxy} is invalid for {url} before.')
                    if request.spider.settings.HTTP_PROXY_FILL_ENABLE:
                        request.proxy = self.get_proxy(request)
                        if request.proxy:
                            self.logger.warn(f'Filling a new proxy {request.proxy} to {url}.')
                    else:
                        self.logger.warn(f'Dropped proxy {proxy} for {url}.')
                        request.proxy = None
                    return request
            request.proxy = gen_proxy(proxy,_type)
            self.logger.debug(f'[{request.spider.name}]Using proxy {request.proxy} '
                              f'for {request.method}-{request.url}')
        else:
            _proxy = None
            while 1:
                _proxy = self.get_proxy(request)
                if _proxy is None:
                    break
                proxy = extract_ip_port(_proxy)
                if self._proxy_invalid(proxy,url):
                    continue
                break
            request.proxy = _proxy
        return request

    def process_response(self,response):
        settings = response.spider.settings
        fakes = settings.HTTP_PROXY_FAKE_STATUS
        domain = parse_url(response.url).netloc
        if not response.spider.settings.HTTP_PROXY_ENABLE:
            return response
        if response.request.proxy and response.status != 200 \
                and response.status not in fakes:
            proxy = extract_ip_port(response.request.proxy)
            if proxy not in self.invalid_pool:
                self.invalid_pool[proxy] = set()
            self.logger.debug(f'Proxy {proxy} is invalid for '
                             f'{domain}.')
            self.invalid_pool[proxy].add(domain)
        elif  response.request.proxy and (response.status == 200
                                          or response.status in fakes):
            proxy = extract_ip_port(response.request.proxy)
            if proxy in self.invalid_pool:
                self.invalid_pool[proxy].discard(domain)
            self.proxy_pool.add(proxy)
        return response

    def get_proxy(self,req):
        http_proxy = req.spider.settings.HTTP_PROXY
        if http_proxy:
            if is_proxy_valid(http_proxy):
                proxy =  gen_proxy(http_proxy,req.down_type)
                return proxy
            elif is_url(http_proxy):
                return http_proxy
            else:
                if not req.spider.settings.HTTP_PROXY_FILL_ENABLE:
                    self.logger.debug(f'Invalid proxy format:{http_proxy}')
                    return
        _proxy = self.get_proxy_by_api(req)
        proxy = gen_proxy(_proxy,req.down_type)
        return proxy


    def get_proxy_by_api(self,request):
        domain = parse_url(request.url).netloc
        def _get_from_pool():
            while self.proxy_pool:
                proxy = self.proxy_pool.pop()
                if proxy not in self.invalid_pool or\
                        (domain not in self.invalid_pool.get(proxy)):
                    return proxy
                else:
                    continue
        proxy = _get_from_pool()
        if not proxy:
            self.logger.debug(f'No proxy in proxy pool.Getting some.')
            while 1:
                spider = request.spider
                req = amipy.Request(spider, spider.settings.HTTP_PROXY_API, delay=0, ignore=True)
                crawler = spider.binding_hub._crawler
                looper = spider.binding_hub.looper
                coro = crawler.requesters[req.down_type].crawl(req)
                resp = looper.run_coroutine(coro)
                if not resp:
                    self.logger.error(f'[{resp.status}]Getting Http proxy by api failed.')
                    continue
                _results = [i.strip() for i in resp.text().split('\n')]
                results =  [is_proxy_valid(i)[0] for i in _results if is_proxy_valid(i)]
                self.proxy_pool.update(results)
                self.logger.debug(f'Got {len(results)} http proxies from HTTP_PROXY_API.')
                proxy = _get_from_pool()
                if not proxy:
                    continue
                break
        return proxy





#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-15
'''
import six
import inspect
from pkgutil import iter_modules
from importlib import import_module

def load_py(path):
    """
    *   load an object from a python module path like:
        'amipy.spider.Spider'
        it will load the Spider Class from spider.py
    *   load a module by given name:
        'amipy.spider'
        it will load module amipy.spider as return result
    """
    try:
        index = path.rindex('.')
    except ValueError:
        obj = import_module(path)
        return obj
    module, name = path[:index], path[index + 1:]
    mod = import_module(module)
    try:
        obj = getattr(mod, name)
    except AttributeError:
        obj = import_module(path)
    return obj

def walk_modules(path):
    """Loads a module and all its submodules from the given module path and
    returns them. If *any* module throws an exception while importing, that
    exception is thrown back.
    Referenced from Scrapy.
    """
    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods

def iter_spider_classes(module):
    """Return an iterator over all spider classes defined in the given module
    that can be instantiated (ie. which have a name)
    """
    from amipy import Spider

    for obj in six.itervalues(vars(module)):
        if inspect.isclass(obj) and \
           issubclass(obj, Spider) and \
           obj.__module__ == module.__name__:
           if getattr(obj, 'name'):
               yield obj
           else:
               from amipy.exceptions import SpiderNameNotSet
               raise SpiderNameNotSet(f'Spider {obj.__name__}'
                                      f' does not have a name.')



#coding:utf-8

import asyncio
from amipy.BaseClass import Crawler,CrawlRequester
from amipy.cmd import _iter_specify_classes
from amipy.middlewares import MiddleWareManager
from amipy.log import getLogger

class WebCrawler(Crawler):

    tasks = []
    mw_manager = None

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super(WebCrawler, cls).__new__(cls)
        return cls._instance

    def __init__(self,settings):
        super(WebCrawler,self).__init__()
        self.settings = settings
        self.logger = getLogger(__name__)
        self.semaphore = asyncio.Semaphore(
            self.settings['project'].CONCURRENCY)
        self._install_requester()

    def _install_requester(self):
        _cls = {}
        _module = self.settings['project'].CRAWLING_REQUESTER_MODULE
        for cls in _iter_specify_classes(_module,CrawlRequester):
            cls._crawler = self
            _cls[cls._down_type]=cls()
            self.logger.debug(f'Installed requester "{cls.__name__}".')
        self.requesters = _cls

    @MiddleWareManager.handle_req
    def convert(self,requests):
        self.logger.debug(f'Received {len(requests)} Requests.')
        tasks = []
        for req in requests:
            coro = self.requesters[req.down_type].crawl(req)
            task = asyncio.ensure_future(coro)
            task.add_done_callback(req.delegate_func)
            tasks.append(task)
        self.tasks.extend(tasks)
        self.logger.debug(f'Converted {len(tasks)} Tasks.')
        return tasks

    @property
    def runing_tasks(self):
        return [i for i in self.tasks if not i.done()]

    @property
    def finished_tasks(self):
        return [i for i in self.tasks if i.done()]







from amipy.log import getLogger

class Scheduler(object):

    def __init__(self,settings):
        self.req_limits = settings.gets('CONCURRENCY')
        self.recv_req = []
        self.waiting = False
        self.spiders = None
        self.logger = getLogger(__name__)
        self.logger.debug('Loaded scheduler.')

    def receive(self,req_queue):
        def any_daemon():
            return any(i.status in ['PAUSE','STOP'] for i in self.spiders)
        if not self.waiting:
            self.logger.debug(f'Requests Queue Size:{req_queue.qsize()}')
        if not req_queue.empty():
            self.waiting = False
            for _ in range(min(self.req_limits,req_queue.qsize())):
                self.recv_req.append(req_queue.get_nowait())
                req_queue.task_done()
            self.logger.debug(f'Left Requests:{req_queue.qsize()}')
        else:
            self.waiting = True
            if all(i.status in ['RUNNING','CLOSE'] for i in self.spiders):
                if self._gather_retry():
                    self.logger.info(f'Start to retry {len(self.recv_req)}'
                                     f' Error and Exception pages.')
                    return
                print('\n* [Done] No Requests to start the crawling.\n')
                raise StopAsyncIteration
            if any_daemon():
                return

    def export(self):
        _export = []
        while self.recv_req:
            _export.append(self.recv_req.pop(0))
        if not self.waiting:
            self.logger.debug(f'Exported {len(_export)} Requests.')
        return _export

    def spiders_monitor(self,spiders):
        self.spiders = spiders
        def not_running():
            return all([i.status in ['STOP','PAUSE'] for i in spiders])
        while not_running():
            for i in spiders:
                if i.status=='STOP' and not i.stopped:
                    self.logger.debug(f'Stopping spider {i.name}.')
                    for req in i.binding_hub.requests._queue:
                        if req.spider.name == i.name:
                            i.binding_hub.requests._queue.remove(req)
                            self.logger.debug(f'Removing request {req}.')
                    i.stopped = True
            continue
        if all(i.status=='CLOSE' for i in spiders):
            self.logger.info('* All spiders closed.')
            raise StopAsyncIteration
        for i in spiders:
            if i.status == 'RESUME':
                self.logger.debug(f'Resuming spider {i.name}.')
                i.resume()
            if i.status == 'RESTART':
                self.logger.debug(f'Restarting spider {i.name}.')
                i.restart()
            if i.status == 'CLOSE':
                self.logger.debug(f'Closing spider {i.name}.')
                i.close(True)

    def _gather_retry(self):
        for i in self.spiders:
            if any(i._retries):
                while i._retries:
                    _req = i._retries.pop(0)
                    self.recv_req.append(_req)
                self.logger.info(f'Got {len(self.recv_req)} retry Requests of {i.name}.')
        return bool(self.recv_req)







#coding:utf-8

from amipy.util.dbhelper import Database

class ModelMetaclass(type):
    def __new__(cls, name, bases, attrs):
        if name=='Model':
            return type.__new__(cls, name, bases, attrs)
        mappings = dict()
        for k, v in attrs.items():
            if isinstance(v, Field):
                mappings[k] = v
        for k in mappings.keys():
            attrs.pop(k)
        attrs['__mappings__'] = mappings
        attrs['__table__'] = name
        return type.__new__(cls, name, bases, attrs)

class Model(dict,metaclass=ModelMetaclass):
    '''Data item model'''

    def __init__(self, **kw):
        super(Model, self).__init__(**kw)
        self._settings = None

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Model' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):
        self[key] = value

    def __init_db__(self):
        self.db = Database(self._settings.DATABASE_SETTINGS)
        self.db.table = self.__table__
        self.db.connect()

    def save(self,format=None):
        fields = []
        _data = {}
        args = []
        for k, v in self.__mappings__.items():
            fields.append(v.name)
            value = getattr(self, k, None)
            args.append(value)
            _data[v.name] = value
        if _data:
            self.db.save(_data,format)
            for k, v in self.__mappings__.items():
                setattr(self,k,None)

    def export(self,path,kind='EXCEL'):
        pass

class Field(object):
    '''The base field of the ORM data model'''
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return '<%s:%s>' % (self.__class__.__name__, self.name)


#coding:utf-8
import os
import shutil
from amipy.exceptions import PathDoesntExist

def check_path(path,make=False):
    if os.path.exists(path):
        return True
    else:
        if make:
            os.makedirs(path)
            return True
        return False

def copy_files(src,dst,ignore_pattern,make=True,render=None,strips='.tpl'):
    if not check_path(src):
        raise PathDoesntExist
    if not check_path(dst,make):
        raise PathDoesntExist
    src_files = os.listdir(src)
    ignore_fun = shutil.ignore_patterns(*ignore_pattern)
    filter_matches = ignore_fun(src,src_files)
    for file in src_files:
        if file in filter_matches:
            continue
        srcname = os.path.join(src,file)
        dstname = os.path.join(dst,file.rstrip(strips))
        if os.path.isdir(srcname):
            copy_files(srcname,dstname,ignore_pattern)
        else:
            shutil.copy2(srcname,dstname)
            if render:
                if file in render:
                    _var = render[file]
                    render_template(srcname,dstname,_var)

def render_template(src,dst,var,encoding='utf-8'):
    with open(src,'r',encoding=encoding) as f:
        tpl = f.read()
    import string
    s = string.Template(tpl).safe_substitute(var)
    with open(dst,'w',encoding=encoding) as f:
        f.write(s)

def get_file_size(path=None,size=None):
    if not size:
        size = os.path.getsize(path)
    elif not path:
        size = size
    k = 1024
    if k>size:
        return f'{size}b'
    elif k**2 > size:
        return f'{round(size/k)}KB'
    elif k**3 > size:
        return f'{round(size/k**2,2)}MB'
    elif k**4 <size:
        return f'{round(size/k**3,2)}GB'




__all__ = ['Spider','Request','SpiderHub','Response','Url','send']

from amipy.spider import Spider
from amipy.request import Request
from amipy.response import Response
from amipy.core.spiderhub import SpiderHub
from amipy.BaseClass import Url
from amipy.util.tools import send
import sys
if sys.version_info < (3, 5):
    print("Amipy requires Python 3.5.x or later version. " )
    sys.exit(1)
del sys





from functools import wraps
from amipy.BaseClass import Middleware
from amipy.cmd import _iter_specify_classes
from amipy.exceptions import DropRequest,DropResponse
from amipy.log import getLogger

class MiddleWareManager(object):

    mw = {}
    req_mw = {}
    resp_mw = {}
    logger = None

    def __init__(self,settings,spiders):
        self._settings = settings
        self.spiders = spiders
        self._attrs = ('mw','resp_mw','req_mw')
        self.logger = getLogger(__name__)
        MiddleWareManager.logger = self.logger
        self.load_middlewares()

    def load_middlewares(self):
        _req_mw = {}
        _resp_mw = {}
        common_req_mw = self._settings['project'].MIDDLEWARE_COMMON_INSTALL['request']
        common_resp_mw = self._settings['project'].MIDDLEWARE_COMMON_INSTALL['response']
        common_both_mw = self._settings['project'].MIDDLEWARE_COMMON_INSTALL['both']
        for spider in self.spiders:
            self.mw[spider.name] = spider.settings.MIDDLEWARE_TO_INSTALL
            common_req_mw.update(self.mw[spider.name]['request'])
            common_resp_mw.update(self.mw[spider.name]['response'])
            common_both_mw.update( self.mw[spider.name]['both'])
            _req_mw[spider.name] = common_req_mw
            _resp_mw[spider.name] = common_resp_mw
            _req_mw[spider.name].update(common_both_mw)
            _resp_mw[spider.name].update(common_both_mw)

        def _load(name_dict):
            mws = {}
            for name,modules in name_dict.items():
                mws[name]=[]
                for i in sorted(modules, key=lambda x: -modules[x]):
                    for j in _iter_specify_classes(i, Middleware):
                        mws[name].append(j())
                        self.logger.debug(f'Loaded middleware [{name}] "{j.__name__}".')
            return mws

        self.req_mw = _load(_req_mw)
        self.resp_mw = _load(_resp_mw)
        [setattr(self.__class__, i,
                 self.__getattribute__(i)) for i in self._attrs]

    @classmethod
    def _process_request(cls,request):
        for i in cls.req_mw[request.spider.name]:
            try:
                request = i.process_request(request)
            except DropRequest:
                # cls.logger.debug(f'{i.__class__.__name__} Dropped '
                #                  f'[{request.spider.name}]-{request.method} {request.url}')
                return None
        return request

    @classmethod
    def _process_response(cls,future):
        response = future.result()
        if not response:
            return None
        if response.request._ignore:
            return response
        for i in cls.resp_mw[response.spider.name]:
            try:
                response = i.process_response(response)
            except DropResponse:
                # cls.logger.debug(f'{i.__class__.__name__} Dropped '
                #                  f'[{response.spider.name}]-[{response.status}] {response.url}')
                return None
        return response

    @classmethod
    def handle_req(cls,func):
        @wraps(func)
        def wrap(_self,requests,*args,**kwargs):
            if not any(requests):return None
            _r = []
            cls.logger.debug(f'Before Middleware handling:{len(requests)} Requests.')
            for req in sorted(requests,key=lambda x:-x.priority):
                if req._ignore:
                    _r.append(req)
                    continue
                request = cls._process_request(req)
                if isinstance(request,list):
                    _r.extend(request)
                else:
                    _r.append(request)
            if not any(_r):return None
            requests = [i for i in _r if i]
            cls.logger.debug(f'After Middleware handling:{len(requests)} Requests.')
            return func(_self,requests,*args,**kwargs)
        return wrap

    @classmethod
    def handle_resp(cls,func):
        @wraps(func)
        def wrap(_self, future, *args, **kwargs):
            response = cls._process_response(future)
            if response is None:
                return
            return func(_self, response, *args, **kwargs)
        return wrap


import time
import re as _re
import w3lib.url as urltool
from amipy.exceptions import NoDataModelBuild
from amipy.BaseClass.orm import Model

class BaseSpider(object):

    name = None
    binding_hub = None
    settings = None
    priority = 0
    urls = []
    my_settings = {}
    whitelist = []
    blacklist = []
    rules = []
    _start_at = time.ctime()
    _pause_at = None
    _resume_at = None
    _restart_at = None
    _stop_at = None
    _close_at = None
    _success = 0
    _fail = 0
    _exc = 0
    urlfilter = None
    respfilter = None
    _hanged = []
    _retries = []
    closed = False
    stopped = False
    _meta = {}
    _models = {}
    _model = None

    @property
    def item(self):
        if self._model:
            return self._model
        else:
            raise NoDataModelBuild('Please build a data model of "item.py" first.')

    @item.setter
    def item(self,model):
        if isinstance(model,Model):
            model._settings = self.settings
            model.__init_db__()
            self._model = model
        else:
            raise TypeError(f'Invalid type,expected "Model" object,got "{type(model)}".')

    def parse(self,response):
        '''Callback function to handle the success response'''
        raise NotImplementedError

    def error(self,response):
        '''Callback function to handle the status code of response that is not 200'''
        pass

    def exception(self,response):
        '''Callback function to handle the exception of response'''
        pass

    def fingerprint(self,response):
        '''indicate the base filter content to differ website pages'''
        return Fingerprint(None,False)

    def send(self,request):
        '''Send a Request to the binding SpiderHub'''
        raise NotImplementedError

    def start_requests(self):
        '''Generate the starting Requests for the spider.'''
        raise NotImplementedError

    def __build_model__(self):
        import os,sys,six,inspect
        from importlib import import_module
        _files = os.listdir(self.settings.PATH)
        if 'item.py' in _files:
            _module = self.settings.PATH.split(os.sep)[-1]
            _name =  '.'.join([_module, 'item'])
            sys.path.append(self.settings.PATH
                            [:self.settings.PATH.rindex(os.sep)] + os.sep)
            _item = import_module(_name)
            for obj in six.itervalues(vars(_item)):
                if inspect.isclass(obj) and \
           issubclass(obj, Model) and obj!=Model:
                    model = obj()
                    model._settings = self.settings
                    model.__init_db__()
                    self._models[model.__table__] = model
            if len(self._models)==1:
                self._model = self._models.popitem()[1]
            else:
                self._model = self._models

    def __gt__(self, other):
        return self.priority < other.priority


class Hub(object):

    spiders = []
    priorities = 0

    def accept(self,request):
        '''accept a Request from a spider.'''
        raise NotImplementedError

    def export(self):
        '''export Request objects from the hub.'''
        raise NotImplementedError

    def takeover(self,spiders):
        '''take over spiders but not binding yet.'''
        raise NotImplementedError

    def delegate(self,future):
        '''a delegate callback function when WebCrawler finished handling a Request.'''
        raise NotImplementedError

    def _binding(self):
        '''bind each spider to the hub'''
        raise NotImplementedError

class Crawler(object):

    def convert(self,requests):
        '''convert a Request list to a Task list'''
        raise NotImplementedError


class Url(object):
    '''an url identifier'''
    def __init__(self,url=None,re=None,domain=None,
                 path=None,params=None,query=None,
                 fragment=None,callback=None,
                 unmatch=None,drop=False,scheme=None,
                 filter=None,fingerprint=None,
                 cookies=None,obey_robots_txt=None,
                 down_type=None,proxy=None,
                 proxy_auth=None):
        self._url = url
        self.scheme = scheme
        self.path = path
        self.params = params
        self.query = query
        self.fragment = fragment
        self.pattern = _re.compile(re) if re else re
        self.domain = domain
        self.drop = drop
        self.cb = callback
        self.unmatch = unmatch
        self.filter = filter
        self.cookies = cookies
        self.fingerprint = fingerprint
        self.obey_robots_txt = obey_robots_txt
        self.down_type = down_type
        self.proxy = proxy
        self.proxy_auth = proxy_auth

    @property
    def url(self):
        if self._url is None:
            return
        return urltool.canonicalize_url(
            urltool.safe_download_url(self._url)
            , encoding='utf-8')

    @url.setter
    def url(self,v):
        from amipy.exceptions import Forbidden
        raise Forbidden('Setting url of an Url is forbidden.')

    def match(self,url):
        _m = None
        if self.pattern:
            _m = self.pattern.findall(url)
        parsed_url = urltool.parse_url(url)
        if (_m and _m[0] == url) or \
            url == self.url or \
            parsed_url.scheme == self.scheme or \
            parsed_url.path == self.path or \
            parsed_url.query == self.query or \
            parsed_url.params == self.params or \
            parsed_url.netloc == self.domain or \
            parsed_url.fragment == self.fragment:
            return url

class Command(object):
    '''A console side commands recognition class'''
    requires_project = False
    parser  = None
    name = None

    def __init__(self,parser,name):
        self.parser = parser
        self.name = name

    def syntax(self):
        '''commands syntax (preferably one-line). Do not include commands name'''
        raise NotImplementedError

    def short_desc(self):
        '''a short description of the command'''
        raise NotImplementedError

    def long_desc(self):
        '''a long description of the command'''
        pass

    def add_options(self,parser):
        pass

    def help(self):
        '''shows the usage of this command.'''
        pass

    def handle(self,settings,opts, args):
        '''handle the command with the args and determine where to go'''
        raise NotImplementedError

    def run(self,settings):
        '''run the command'''
        raise NotImplementedError

class SpiderClientCommand(object):
    '''the base class of spider server commands tools at client side.'''

    def parse(self,cmdname,args,spiders):
        '''return the cmd parsed result'''
        raise NotImplementedError


class CrawlRequester(object):
    '''the base class of sending Requests generated by different kinds of spiders'''
    _down_type = None
    _crawler = None

    def crawl(self,request):
        '''defines how to deal with Requests'''
        raise NotImplementedError

class Middleware(object):
    '''identifies the middlewares'''

    def process_request(self,request):
        return request

    def process_response(self,response):
        return response

class Fingerprint(object):

    def __init__(self,text,precise=True):
        self.text = text
        self.precise = precise







import os
import asyncio
from amipy import Response
from amipy.BaseClass import CrawlRequester
from amipy.util.http import send_async_http
from amipy.util.file import get_file_size
from amipy.log import getLogger

class MediaRequester(CrawlRequester):

    _down_type = 'media'

    logger = getLogger(__name__)

    async def crawl(self,request):
        delay = request.delay
        url = request.url
        session = request.spider.session
        proxy = request.proxy
        buffer = request.spider.settings.DEFAULT_DOWNLOAD_BUFFER
        path = os.path.normpath(request.save_path)
        if not os.path.exists(os.path.dirname(path)):
            self.logger.error(f'No path:{os.path.dirname(path)}.')
            return
        name = os.path.basename(path)
        try:
            self.logger.info(f'Downloading {name}.')
            async with self._crawler.semaphore:
                resp = await send_async_http(   session,
                                                request.method,
                                                url,
                                                path=path,
                                                retries=request.retry,
                                                timeout=request.timeout,
                                                proxies=proxy,
                                                buffer=buffer
                                            )
                if resp is None:
                    return
                body = resp['body']
                exception = resp['exception']
                if exception and body != True:
                    return Response(url, status=-1, request=request, exc=exception)
            await asyncio.sleep(delay)
            size = get_file_size(size=int(resp['size']))
            self.logger.info(f'Finished downloading:[{name} {size}]')
            return
        except asyncio.CancelledError:
            print(f'Task "{request}" canceled.')
            return Response(url, status=0, request=request)
        except Exception as e:
            return Response(url, status=-1, request=request, exc=e.__class__())



import w3lib.url  as urltool
from amipy.middlewares import Middleware
from amipy.exceptions import DropRequest
from amipy import Url

class UrlListHandleMiddleware(Middleware):

    def process_request(self,request):
        spider = request.spider
        url = request.url
        blacklist = spider.blacklist
        whitelist = spider.whitelist
        if self._is_in(url,blacklist):
            raise DropRequest
        if any(whitelist) and not \
                self._is_in(url,whitelist):
            raise DropRequest
        return request

    def _is_in(self,_url,_list):

        def _tran(url):
            return urltool.canonicalize_url(
            urltool.safe_download_url(url), encoding='utf-8')

        for url in _list:
            if isinstance(url,Url):
                if url.match(_url):
                    return True
            elif _tran(url) == _url:
                return True



import threading, time
from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):

    def parse(self,cmdname,args,spiders):
        argv = args[0]
        urls = args[1:]
        if argv == 'spiders':
            return '\r\n Not a valid usage of command: restart <spider name> [restart url] [restart url]'
        d = {i.name:i for i in spiders}
        if argv in d:
            spider = d[argv]
            if spider.status == 'RUNNING':
                return f"* Spider {argv} is running,it can not restart. "
            elif spider.status == 'STOP':
                lock = threading.Lock()
                lock.acquire()
                spider.status = 'RESTART'
                spider._meta['restart_urls'] = urls
                spider._restart_at = time.ctime()
                lock.release()
                return f'\r\n* Spider {argv} restarted at {spider._restart_at} successfully.'
            elif spider.status == 'CLOSE':
                return f'* Spider {argv} is closed.'
            elif spider.status == 'PAUSE':
                return f'* Spider {argv} is paused.Using "resume" command to resume it.'
            else:
                return f'* Invalid resuming status "{spider.status}" for a spider.'
        else:
            return f'* No spider "{argv}" in the project.'






import threading, time
from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):

    def parse(self,cmdname,args,spiders):
        argv = args[0]
        if argv == 'spiders':
            return '\r\n Not a valid usage of command: stop <spider name>'
        d = {i.name:i for i in spiders}
        if argv in d:
            spider = d[argv]
            if spider.status == 'STOP':
                return f"* Spider {argv} hab been stopped at {spider._stop_at} already. "
            else:
                lock = threading.Lock()
                lock.acquire()
                spider.status = 'STOP'
                spider._meta['restart_urls'] = []
                spider._stop_at = time.ctime()
                lock.release()
                return f'* {spider._stop_at} Spider {argv} stopped successfully. '
        else:
            return f'* No spider "{argv}" in the project.'




import threading, time
from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):

    def parse(self,cmdname,args,spiders):
        argv = args[0]
        if argv == 'spiders':
            return '\r\n Not a valid usage of command: pause <spider name>'
        d = {i.name:i for i in spiders}
        if argv in d:
            spider = d[argv]
            if spider.status == 'PAUSE':
                return f"* Spider {argv} hab been paused at {spider._pause_at} already. "
            elif spider.status == 'RUNNING' or \
                 spider.status == 'RESUME':
                lock = threading.Lock()
                lock.acquire()
                spider.status = 'PAUSE'
                spider._pause_at = time.ctime()
                lock.release()
                return f'* {spider._pause_at} Spider {argv} paused successfully. '
            else:
                return f'* Invalid pausing status "{spider.status}" for a spider. '
        else:
            return f'* No spider "{argv}" in the project.'



#coding:utf-8
import os
import amipy
import aiohttp
import bloompy
from amipy.util.filter import _generate_filter
from amipy.core.spiderhub import SpiderHub
from amipy.BaseClass import BaseSpider
from amipy.log import getLogger

class Spider(BaseSpider):

    def __init__(self,*args,**kwargs):
        super(BaseSpider, self).__init__(*args,**kwargs)
        self.status = 'CREATED'
        self.requests = []
        self.session = None
        self.logger = getLogger(__name__)
        self._load_settings()
        self._load_filters()
        self._init_session()

    def __str__(self):
        return f'<{self.__class__.__name__}({self.name}) object' \
               f' at {hex(id(self))} success:{self._success} ' \
               f'fail:{self._fail} exc:{self._exc} {self.status}>'

    def start_requests(self,urls=None):
        urls = self.urls if not bool(urls) else urls
        return [amipy.Request(self, url, ignore=True) for url in urls]

    def _load_settings(self):
        from inspect import ismodule
        if not ismodule(self.settings):
            self.settings = type('s', (), self.my_settings)
        if self.my_settings:
            for key,value in self.my_settings.items():
                if key=='NAME':continue
                setattr(self.settings,key,value)

    def _load_filters(self):
        u_path = self.settings.BLOOMFILTER_URL_LOAD_PATH
        h_path = self.settings.BLOOMFILTER_HTML_LOAD_PATH
        if self.settings.BLOOMFILTER_URL_ON:
            if os.path.exists(u_path):
                self.urlfilter = bloompy.get_filter_fromfile(u_path)
            else:
                self.urlfilter = _generate_filter(self.settings)
            self.logger.debug(f'Loaded [{self.name}] urlfilter,'
                          f'count:{self.urlfilter.count},'
                          f'capacity:{self.urlfilter.capacity},'
                          f'error rate:{self.urlfilter.error_rate}')
        if self.settings.BLOOMFILTER_HTML_ON:
            if os.path.exists(h_path):
                self.respfilter = bloompy.get_filter_fromfile(h_path)
            else:
                self.respfilter = _generate_filter(self.settings)
            self.logger.debug(f'Loaded [{self.name}] respfilter,'
                              f'count:{self.respfilter.count},'
                              f'capacity:{self.respfilter.capacity},'
                              f'error rate:{self.respfilter.error_rate}')

    def _init_session(self):
        _safe = self.settings.SPIDER_COOKIES_UNSAFE_MODE
        path = self.settings.SPIDER_COOKIES_LOAD_PATH
        _c_cookies = self.settings.SPIDER_COOKIES_CUSTOM
        jar = aiohttp.CookieJar(unsafe=_safe)
        if _c_cookies:
            cookies = _c_cookies
        else:
            cookies = None
        self.conn = aiohttp.TCPConnector(limit=self.settings.CONCURRENCY)
        self.session = aiohttp.ClientSession(connector=self.conn,
                                             cookies=cookies,
                                             cookie_jar=jar)
        if path:
            if os.path.exists(path):
                try:
                    self.session.cookie_jar.load(path)
                    if cookies:
                        self.session.cookie_jar.update_cookies(cookies)
                except:
                    return
                self.logger.debug(f'Loaded [{self.name}] cookie jar.')

    @property
    def cookies(self):
        return self.session.cookie_jar._cookies

    @cookies.setter
    def cookies(self,_cookies):
        if isinstance(_cookies,dict):
            self.session.cookie_jar.update_cookies(_cookies)
        else:
            import warnings
            warnings.warn(f'Not a valid cookie type,expected Dict,'
                          f'got:{type(_cookies).__name__}')

    def send(self, request):
        if not isinstance(request, amipy.Request):
            raise TypeError(f'not a valid Request to send,'
                            f'got "{type(request).__name__}".')
        if not isinstance(self.binding_hub,SpiderHub):
            raise TypeError('Not a valid binging SpiderHub for Spider %s,got "%s".'
                            %(self.name, type(self.binding_hub).__name__))
        self.binding_hub.requests.put_nowait(request)

    def save_cookies(self):
        path = self.settings.SPIDER_COOKIES_SAVE_PATH
        if path:
            try:
                self.session.cookie_jar.save(path)
            except:
                return
            self.logger.debug(f'Save cookies of "{self.name}" succeed.')

    def save_records(self):
        if self.settings.BLOOMFILTER_URL_ON:
            path = self.settings.BLOOMFILTER_URL_SAVE_PATH
            if path:
                self.urlfilter.tofile(path)
                self.logger.debug(f'Save urlfilter of "{self.name}" succeed.')
        if self.settings.BLOOMFILTER_HTML_ON:
            path = self.settings.BLOOMFILTER_HTML_SAVE_PATH
            if path:
                self.respfilter.tofile(path)
                self.logger.debug(f'Save respfilter of "{self.name}" succeed.')

    def close(self,save=True):
        if self.closed:
            return
        self.status = 'CLOSE'
        if save:
            self.save_cookies()
            self.save_records()
        self.conn.close()
        self.closed = True

    def resume(self):
        self.status = 'RUNNING'
        while self._hanged:
            a = self._hanged.pop(0)
            if a:
                self.logger.debug(f'RESUME {self.name} {a}')
                self.binding_hub.requests.put_nowait(a)

    def restart(self):
        self.status = 'RUNNING'
        self.stopped = False
        if self._meta.get('restart_urls'):
            urls = self._meta['restart_urls']
        else:
            urls = []
        req = self.start_requests(urls)
        self.logger.debug(f'RESTART {self.name} {req}')
        [self.binding_hub.requests.put_nowait(i) for i in req if i]


import amipy
from w3lib.url import parse_url
from amipy.middlewares import Middleware
from amipy.log import getLogger
from urllib import robotparser
from amipy.exceptions import DropRequest

class RobotstxtMiddleware(Middleware):

    def __init__(self):
        self.rparser = {}
        self.rubbish = set()
        self.logger = getLogger(__name__)

    def process_request(self,request):
        spider = request.spider
        if  not request.obey_robots_txt:
            if not spider.settings.ROBOTS_TXT_OBEY:
                return request
        _purl = parse_url(request.url)
        netloc = _purl.netloc
        if not netloc in self.rparser :
            if netloc in self.rubbish:
                return request
            robots_url = f"{_purl.scheme}://{netloc}/robots.txt"
            req = amipy.Request(spider, robots_url)
            crawler = spider.binding_hub._crawler
            looper = spider.binding_hub.looper
            coro = crawler.requesters[req.down_type].crawl(req)
            resp = looper.run_coroutine(coro)
            if resp.status != 200:
                self.logger.debug(f'[{resp.status}] There is no robots.txt for "{netloc}".')
                self.rubbish.add(netloc)
                return request
            else:
                self.logger.debug(f'[{resp.status}] Found robots.txt for "{netloc}".')
                _parser = robotparser.RobotFileParser(robots_url)
                _parser.parse(resp.text().splitlines())
                self.rparser[netloc] = _parser
        else:
            _parser = self.rparser[netloc]
        ua = spider.settings.ROBOTS_USER_AGENT
        if _parser.can_fetch(ua,request.url):
            return request
        else:
            self.logger.debug(f'Forbidden by robots.txt of "{netloc}".'
                              f'Request:{request}')
            raise DropRequest









from amipy.middlewares import Middleware
from asyncio import TimeoutError
from amipy.log import getLogger
from aiohttp.client_exceptions import ServerDisconnectedError,\
    ServerConnectionError,ClientOSError,ClientConnectionError

exceptions = [TimeoutError,ServerConnectionError,
              ServerDisconnectedError,ClientConnectionError,
              ClientOSError]

class RetryPagesMiddleware(Middleware):

    logger = getLogger(__name__)

    def process_response(self,response):
        spider = response.spider
        codes = spider.settings.REQUESTS_ERROR_RETRY_STATUS
        if spider.settings.REQUESTS_FAIL_RETRY_ENABLE:
            if response.status != 200:
                if (response.status in codes
                    and response.status != -1) or \
                    (response.exception.__class__ in exceptions):
                    _tried = response.request._tried
                    if _tried > spider.settings.REQUESTS_FAIL_RETRY_DEPTH:
                        return response
                    response.request._tried = _tried + 1
                    response.request.proxy=None
                    self.logger.debug(f'{response.request} scheduled to retry.Tried:{_tried}')
                    spider._retries.append(response.request)
        return response






from amipy.core.workstation import WorkStation
from amipy.BaseClass import Command
from optparse import OptionGroup

class AnyNameYouWant(Command):

    requires_project = True

    def handle(self,settings,opts,args):
        excludes = None
        if opts.exclude:
            import re,string
            excludes = re.sub('[%s ]'%string.punctuation,'',opts.exclude).split(',')
        works = WorkStation(settings)
        works.work(excludes=excludes)

    def add_options(self,parser):
        group = OptionGroup(parser, 'Additional options')
        group.add_option("-e", "--exclude", metavar="SPIDER", dest='exclude',
                         help="run all the spiders inside the project exclude the specified spiders.")
        parser.add_option_group(group)

    @classmethod
    def short_desc(self):
        return 'start to run all the spiders inside the project.'

    def help(self):
        pass

    def syntax(self):
        return ' <project name> [options] [args] '




import re,string
from amipy.BaseClass import Command
from amipy.exceptions import CommandUsageError
from amipy.core.workstation import WorkStation

class AnyNameYouWant(Command):

    requires_project = True

    def handle(self, settings, opts, args):
        if not all(args):
            raise CommandUsageError(self,self.parser)
        spider_names = [re.sub('[%s ]'%string.punctuation,'',i) for i in args]
        works = WorkStation(settings)
        works.work(spider_names=spider_names)

    @classmethod
    def short_desc(self):
        return 'run a specified spider by a given name.'

    def help(self):
        pass

    def syntax(self):
        return ' <spider name> [options] args '




from amipy.exceptions import DropRequest,DropResponse
from amipy.middlewares import Middleware
from amipy.log import getLogger
from amipy.util.filter import _to_feature

class ServerCtrlMiddleware(Middleware):

    logger = getLogger(__name__)

    def process_request(self,request):
        s = request.spider
        if s.status == 'PAUSE':
            self.logger.debug(f'PAUSE {s.name} {request}')
            s._hanged.append(request)
            if s.urlfilter:
                s.urlfilter.delete(_to_feature(request))
            raise DropRequest
        elif s.status == 'RUNNING':
            return request
        elif s.status in ['STOP','CLOSE']:
            if s.urlfilter:
                s.urlfilter.delete(_to_feature(request))
            self.logger.debug(f'STOP/CLOSE {s.name} {request}')
            raise DropRequest

    def process_response(self,response):
        spider = response.spider
        if spider.status =='STOP':
            raise DropResponse
        return response



from amipy.middlewares import Middleware

class HeadersMiddleware(Middleware):

    def process_request(self,request):
        spider = request.spider
        headers = request.headers
        _header_cookies = spider.settings.SPIDER_COOKIES_FOR_HEADERS \
        if getattr(spider.settings,'SPIDER_COOKIES_FOR_HEADERS',False) else None
        if headers is None:
            _headers = spider.settings.REQUEST_HEADERS
            request.headers = _headers
        if _header_cookies:
            request.headers['Cookie'] = _header_cookies
        if request.headers.get('Cookie',None) or request.headers.get('cookie',None):
            _c = request.headers.get('Cookie','')
            _c = _c if _c else request.headers.get('cookie','')
            _c = self._parse_cookie(_c)
            spider.session.cookie_jar.update_cookies(_c)
        return request

    def _parse_cookie(self,cookie_str):
        c_dict = {}
        for  i in cookie_str.split(';'):
            k,v = i.split('=')
            c_dict.update({k.strip():v.strip()})
        return c_dict


#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-15
'''
import time
from amipy.core.serverthread import SpiderServer
from amipy.core.spiderhub import SpiderHub
from amipy.core.loader import SpiderLoader
from amipy.core.crawler import WebCrawler
from amipy.core.looper import Looper
from amipy.core.scheduler import Scheduler
from amipy.middlewares import MiddleWareManager
from amipy.log import getLogger
from amipy.util.time import time_to_date


class WorkStation(object):

    def __init__(self,settings):
        self.settings = settings
        self.spider_loader = SpiderLoader(settings)
        self.spiders = self.spider_loader.load_all_spiders()
        self.crawler = WebCrawler(settings)
        self.scheduler = Scheduler(settings)
        self.looper = Looper()
        self.spider_hub = SpiderHub(settings,self.crawler)
        self.logger = getLogger(__name__)

        # self.data_processor = DataProcessor(settings)

    def _print_tips(self,got=True):
        print(f'* Amipy - project : {self.settings["project"].PROJECT_NAME}')
        if got:
            date = time_to_date(int(time.time()))
            print(f'* Running at {date}')
            print(f'* Spiders inside the project: {[i.name for i in self.spiders]}')
        else:
            print('* No spiders inside the project yet.Try to create one!')
            print('* You can create a spider by using commands like:\n')
            print('>> amipy cspider myspider\n')
            print('* Then you will see a directory named "myspider" ')
            print(f'* under the "spiders" folder of the project "{self.settings["project"].PROJECT_NAME}".')
            print('* What you need to do is "edit the spider.py" as you want.')
            self._close()

    def work(self,excludes=None,spider_names=None):
        self.logger.debug('Workstation running.')
        try:
            if excludes:
                spiders = [i for i in self.spiders if i.name not in excludes]
            elif spider_names:
                spiders = [i for i in self.spiders if i.name in spider_names]
                if not spiders:
                    print(f'* Amipy - project : {self.settings["project"].PROJECT_NAME}')
                    print(f'* No spider named "{spider_names}" inside the project.')
                    return
            else:
                spiders = self.spiders
            if not spiders:
                self._print_tips(False)
                return
            else:
                self._print_tips()
                [i.__build_model__() for i in spiders]
                print(f'* Current running spiders: {[i.name for i in  spiders]}')
                self.mw_manager = MiddleWareManager(self.settings, spiders)
                if self.settings['project'].SPIDER_SERVER_ENABLE:
                    self.server = SpiderServer(self.settings,spiders)
                    self.server.start()
                    self.logger.debug('SpiderServer started.')
                else:
                    print('* Press Ctrl+C to stop the crawling.\n')
            self.spider_hub.takeover(spiders)
            self.spider_hub.start(self.looper)
            while 1:
                self.scheduler.spiders_monitor(spiders)
                self.scheduler.receive(self.spider_hub.requests)
                tasks = self.crawler.convert(self.scheduler.export())
                self.looper.run_tasks(tasks)
        except (StopAsyncIteration,KeyboardInterrupt):
            self._close()
            self.logger.info('Amipy had been shut down.')

    def _close(self):
        for i in self.spiders:
            if not i.closed:
                i.close()


#coding:utf-8
from asyncio import PriorityQueue,Queue,LifoQueue

class FIFO(Queue):

    def __init__(self,maxsize=0):
        super(FIFO, self).__init__(maxsize)

class LIFO(LifoQueue):

    def __init__(self,maxsize=0):
        super(LIFO, self).__init__(maxsize)

class Priority(PriorityQueue):

    def __init__(self,maxsize=0):
        super(PriorityQueue, self).__init__(maxsize)





import asyncio
from amipy import Response
from amipy.BaseClass import CrawlRequester
from amipy.util.http import send_async_http

class CommonRequester(CrawlRequester):

    _down_type = 'text/html'

    async def crawl(self,request):
        delay = request.delay
        url = request.url
        session = request.session
        try:
            async with self._crawler.semaphore:
                resp = await send_async_http(session,
                                            request.method,
                                            url,
                                            timeout=request.timeout,
                                            headers=request.headers,
                                            params=request.params,
                                            data=request.data,
                                            retries=request.retry,
                                            proxy=request.proxy)
                _resp = resp['resp']
                body = resp['body']
                exception = resp['exception']
                if exception:
                    return Response(url, status=-1, request=request, exc=exception)
            await asyncio.sleep(delay)
            return Response(url,request=request,body=body,_resp=_resp)
        except asyncio.CancelledError:
            print(f'Task "{request}" canceled.')
            return Response(url, status=0, request=request)
        except Exception as e:
            return Response(url, status=-1, request=request, exc=e.__class__())


# coding:utf-8

import pymongo
import logging
from inspect  import isfunction

logger = logging.getLogger()

CON_MAP = {
    '=':'$eq',
    '<':'$lt',
    '<=':'$lte',
    '>':'$gt',
    '>=':'$gte',
    '!=':'$ne',
}

class Database(object):
    """
    对MongoDB数据库对象的封装，提供更高功能的接口使用
    """
    def __init__(self,settings):
        """
        初始化
        :param settings: 数据库设置
        """
        self.host   = settings['host']
        self.port   = settings['port']
        self.user   = settings['user']
        self.passwd = settings['password']
        self.db     = settings['database']
        self.connected = False
        self.conn    = None
        self.handler = None
        self.table   = None

    def connect(self):
        """
        连接MongoDB
        """
        if self.user and self.passwd:
            self.conn = pymongo.MongoClient(self.host, self.port,username=self.user,password=self.passwd)
        else:
            self.conn = pymongo.MongoClient(self.host, self.port)
        self.handler = self.conn[self.db]
        self.connected = True

    def close(self):
        """
        关闭数据库连接
        """
        self.conn.close()
        self.conn=None

    def use_db(self,dbname):
        """
        连接数据库后使用名为dbname的数据库
        :param dbname: 要使用的数据库
        """
        self.handler = self.conn[dbname]

    def save(self,data,tname=None,format=None):
        """
        保存数据到数据集
        :param data: 要保存的数据,{}类型或 [{},{}..]类型
        :param tname: 数据集(collection)
        :param format:对数据进行格式化的函数，可以根据数据结构自定义
        """
        table = tname if tname else self.table
        format = None if not isfunction(format) else format
        if not table:
            raise Exception('No table or data collection specified by tname.')
        if isinstance(data,dict):
            data = format(data) if format else data
            self.handler[table].insert(data)
        elif isinstance(data,list):
            for i in data:
                if isinstance(i,dict):
                    i = format(i) if format else i
                    self.handler[table].insert(i)
                else:
                    raise TypeError('Expected a dict type value inside the list,%s type received.' % type(data))
        else:
            raise TypeError('Expected a [{},{}..] or {} type data,%s type received.' % type(data))

    def select(self,condition,tname=None,sort=None):
        """
        条件查询数据库得到一个数据列表
        :param condition: 查询条件
        :param tname: 要查询的数据集合名
        :param sort: 排序规则，MongoDB标准，使用dict类型
        :return: 返回查询结果 [{},{},..] 类型
        """
        table = tname if tname else self.table
        if not isinstance(condition,dict):
            raise TypeError('condition is not a valid dict type param.')
        else:
            try:
                conditions = self.__gen_mapped_condition(condition)
                if sort and isinstance(sort,dict):
                    res = self.handler[table].find(condition).sort(list(sort.items()))
                else:
                    res = self.handler[table].find(conditions)
                data = list(res)
                return data if data else []
            except Exception as e:
                logger.error('Error class : %s , msg : %s ' % (e.__class__, e))
                return

    def delete(self,condition,tname=None):
        """
        删除数据库符合条件的数据条目
        :param condition: 删除条件  dict类型
        :param tname: 要删除数据所在的数据集合名
        """
        if not condition: return
        conditions = self.__gen_mapped_condition(condition)
        table = tname if tname else self.table
        if not isinstance(condition,dict):
            raise TypeError('condition is not a valid dict type param.')
        self.handler[table].delete_many(conditions)

    def update(self,condition,data,tname=None):
        """
        按照条件更新数据库数据
        :param condition: 查询条件 dict类型
        :param data: 更新数据 dict类型
        :param tname: 要更新的数据所在的数据集合名
        """
        table = tname if tname else self.table
        if not data :return
        if not isinstance(condition, dict) and not isinstance(data,dict):
            raise TypeError('Params (condition and data) should both be the dict type.')
        conditions= self.__gen_mapped_condition(condition)
        self.handler[table].update(conditions,{'$set':data},False,True )

    def all(self,tname=None):
        """
        返回MongoDB数据库某个集合的所有数据
        :param tname: 数据集合名
        :return: 所有该集合的数据 格式:[{},{},..]
        """
        table = tname if tname else self.table
        data = list(self.handler[table].find())
        return data

    def __gen_mapped_condition(self,condition):
        """
        MongoDB与sql语句的条件查询映射,使其符合MongoDB语法
        如：查询条件为 {'score':{'<':0}}
        则将其映射为   {'score':{'$lt':0}}
        符合MongoDB的查询语法
        :param condition:查询条件 dict类型
        :return: 映射结果 dict类型
        """
        for key in condition:
            if isinstance(condition[key], dict):
                t = condition[key]
                operator = list(t.keys())[0]
                value = t[operator]
                o = CON_MAP[operator]
                condition[key].pop(operator)
                condition[key][o] = value
        return condition




from amipy.BaseClass import Command
from amipy.core.loader import SpiderLoader

class AnyNameYouWant(Command):

    requires_project = True

    def handle(self,settings,opts, args):
        loader = SpiderLoader(settings)
        spiders = loader.load_all_spiders()
        print(f'\n* Active Amipy project:{settings["project"].PROJECT_NAME}')
        print(f'* Found Spiders:{len(spiders)}')
        for _,i in enumerate(spiders):
            print('*','-'*50)
            print(f'* {_+1}. Name:{i.name}   Class:{i.__class__.__name__}  '
                  f'Priority:{i.priority}')
            i.close(save=False)

    @classmethod
    def short_desc(self):
        return 'list all the valid spiders inside the project.'

    def help(self):
        pass

    def syntax(self):
        return '<options> [args]'



#coding:utf-8


class SpiderNameNotSet(Exception):
    '''Indicates that spider's unique name is not set yet'''

class SpiderAlreadyExists(Exception):
    '''Indicates that the spider already exists'''

class InvalidDirPath(Exception):
    '''Indicates that the path is invalid'''

class PathDoesntExist(Exception):
    '''Means that the path of template files does not exist'''

class NotValidJsonContent(Exception):
    '''Indicates the content to JSON is not a valid format'''

class DropRequest(Exception):
    '''the Request needs to ignore,then drop it'''

class DropResponse(Exception):
    '''the Response needs to ignore,then drop it'''

class Forbidden(Exception):
    '''Amipy forbidden options.'''

class FilePathError(Exception):
    '''No existed path for file downloading'''

class NoDataModelBuild(Exception):
    '''There is no data model for saving build.'''

class CommandUsageError(Exception):
    '''a command usage error'''
    def __init__(self,cmd,parser):
        self.cmd = cmd
        self.parser = parser




#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-15
'''

from amipy.exceptions import CommandUsageError
from amipy.BaseClass import Command
from amipy.util.file import copy_files
from optparse import OptionGroup

class AnyNameYouWant(Command):

    requires_project = False

    def add_options(self,parser):
        group = OptionGroup(parser,'Additional options')
        group.add_option("-d","--dir", metavar="DIR",dest='dir',
                         help="specify the project directory.It must be a absolutely path.")
        parser.add_option_group(group)

    @classmethod
    def short_desc(self):
        return 'create a new crawling project.'

    def handle(self,settings,opts,args):
        import os,amipy,re
        if opts.dir:
            if os.path.isdir(opts.dir):
                copy_path = opts.dir
            else:
                raise CommandUsageError(self,self.parser)
        else:
            copy_path = os.getcwd()
        try:
            project_name = re.sub('[^a-zA-Z\d]','',args[0])
        except:
            raise CommandUsageError(self,self.parser)
        destination = os.path.join(copy_path,project_name)
        src = os.path.join(amipy.__path__[0], 'templates.project'.replace('.', os.sep))
        render_vars = {'project_name':project_name,'project_path':destination}
        render_files = {
            'settings.py.tpl':render_vars
        }
        copy_files(src,destination,('.pyc$','.pyd$'),render=render_files)
        print('* Created a project "%s" at "%s".'%(project_name,destination))
        print('Now you can create a spider then edit it to crawl some websites by:\n')
        print('* check inside the project directory,e.g.')
        print('>> cd %s\n'%project_name)
        print('* Create a crawling spider,e.g.')
        print('>> amipy cspider myspider \n')
        print('* Open the spider script and edit it then run it,e.g.')
        print('>> amipy runspider myspider\n')
        print('Go and try!')

    def syntax(self):
        return '<project name> [options] [args] '




from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):


    def parse(self,cmdnae,args,spiders):
        prompt = '''All commands of the Spider-Client:
        \r\n* Usage:
        \r\n  <command> [spider name]
        \r\n* Available commands:
        \r\n       show spiders         show all running spiders and their conditions.
        \r\n       list                 list a general situation of all spiders.
        \r\n       echo                 echo a running spider and its attributes.
        \r\n       pause                pause a running spider by a give name.
        \r\n       stop                 stop a running/paused spider by a give name.
        \r\n       close                close a spider by a give name.
        \r\n       restart              restart a stopped spider by a give name.
        \r\n       resume               resume a paused spider by a give name.
        \r\n       quit                 quit the Spider-Client.
        \r\n       help                 show all the available commands usage.
        '''
        return prompt





import re
import os
from amipy.util.filter import _to_md5,_to_feature
from amipy.middlewares import Middleware
from amipy.exceptions import DropRequest,DropResponse
from amipy.BaseClass import Fingerprint
from bs4 import BeautifulSoup as bs

class CrawlFilterMiddleware(Middleware):

    def process_request(self,request):
        spider = request.spider
        url = request.url
        _flag = self._rules_effect(url,spider)
        if _flag is None:
            if  not request.filter:
                return request
        elif not _flag:
            return request
        _filter = spider.urlfilter
        _feature = _to_feature(request)
        if _feature in _filter:
            raise DropRequest
        else:
            _filter.add(_feature)
        return request

    def _rules_effect(self,url,spider,mode=0):
        rules = spider.rules
        for URL in rules:
            if URL.match(url):
                if mode==0:
                    if URL.filter != None:
                        return bool(URL.filter)
                else:
                    URL_FP =  URL.fingerprint
                    if isinstance(URL_FP,bool):
                        return URL_FP
                    if URL_FP != None and \
                        callable(getattr(spider,URL_FP,False)):
                        return getattr(spider,URL_FP)
        return None

    def process_response(self,response):
        url = response.url
        spider = response.spider
        if response.status != 200 and spider.urlfilter:
            spider.urlfilter.delete(_to_feature(response.request))
        if response.status == -1:
            return response
        self.spider = spider
        _flag = self._rules_effect(url,spider,1)
        if _flag is None:
            if not response.resp_filter:
                return response
        elif not _flag:
            return response
        if callable(response.fingerprint):
            _func = response.fingerprint if \
                not callable(_flag) else _flag
        elif isinstance(response.fingerprint,str) and \
                callable(getattr(spider,response.fingerprint)):
            _func = getattr(spider,response.fingerprint)
        else:
            raise ValueError('Not a valid fingerprint.')
        fingerprint = _func(response)
        if fingerprint.text is None or \
                not isinstance(fingerprint,Fingerprint):
            _fingerprint = response.read()
        else:
            _fingerprint = fingerprint.text
        if fingerprint.precise:
            _feature = _to_md5(_fingerprint)
        else:
            _feature = self._to_analyse(_fingerprint,
                                        spider.settings.BLOOMFILTER_HTML_EXTRACTS)
        if _feature in spider.respfilter:
            raise DropResponse
        else:
            spider.respfilter.add(_feature)
        return response

    def _to_analyse(self,fingerprint,extract_list):
        if fingerprint is None:
            raise DropResponse
        if len(fingerprint)<180:
            return _to_md5(fingerprint)
        html = bs(fingerprint,'lxml')
        [i.extract() for j in extract_list  for i in html(j) ]
        _text = html.body.text
        text = re.sub('\n','',_text).replace('\r','')
        if len(text)<180:
            return _to_md5(fingerprint)
        lines_content = _text.splitlines()
        res_dict = self._extract_content(lines_content)
        if not res_dict:
            return _to_md5(text)
        else:
            keys = sorted(res_dict.keys(),key=lambda x:-x)[:2]
            texts =''.join([res_dict[i] for i in keys])
            return _to_md5(texts)

    def _extract_content(self,lines_content):
        gap = self.spider.settings.BLOOMFILTER_HTML_GAP
        threshold = self.spider.settings.BLOOMFILTER_HTML_THRESHOLD
        density = self.spider.settings.BLOOMFILTER_HTML_DENSITY
        results = {}
        comobo_num = 0
        combo_len = 0
        combo_null = 0
        combo_text = ''
        pre_len = 0
        for i in lines_content:
            if i.strip():
                pre_len = len(i)
                comobo_num += 1
                combo_null = 0
                combo_len += pre_len
                combo_text = combo_text + i + os.linesep
                if len(lines_content) == 1 and pre_len >= density * threshold:
                    results[pre_len] = combo_text
            else:
                combo_null += 1
                if pre_len:
                    if combo_null > gap:
                        if combo_len >= density * threshold \
                                and comobo_num >= threshold:
                            results[combo_len] = combo_text
                    else:
                        continue
                comobo_num = 0
                combo_len = 0 if combo_null > gap else combo_len
                pre_len = 0
                combo_text = '' if combo_null > gap else combo_text
        return results


#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-15
'''
import os
import sys
import six
import optparse
from inspect import isclass
from amipy.config import Settings
from amipy.BaseClass import Command
from amipy.util.load import load_py,walk_modules
from amipy.log import install_logger

APSMP = 'AMICO_PROJECT_SETTINGS_PATH'

def _iter_specify_classes(module_name, cmdcls=Command):
    for module in walk_modules(module_name):
        for obj in six.itervalues(vars(module)):
            if isclass(obj) and issubclass(obj, cmdcls) and \
                    obj.__module__ == module.__name__ and \
                    not obj == cmdcls:
                yield obj

def _get_commands_from_module(module, inproject):
    cmds = {}
    for cmd in _iter_specify_classes(module):
        if inproject or not cmd.requires_project:
            cmdname = cmd.__module__.split('.')[-1]
            cmds[cmdname] = cmd
    return cmds

def _get_all_commands(settings, inproject):
    cmds = _get_commands_from_module('amipy.commands', inproject)
    try:
        cmds_module = settings['project'].COMMANDS_NEW_MODULE
    except:
        pass
    else:
        if cmds_module:
            cmds.update(_get_commands_from_module(cmds_module, inproject))
    return cmds

def _pop_command_name(args):
    for i in enumerate(args[1:]):
        if not i[1].startswith('-'):
            del args[i[0]+1]
            return i[1]

def _init_project(setPath=False):
    cwd = os.getcwd()
    cwd_files = os.listdir(cwd)
    if 'settings.py' in cwd_files:
        cwd_module = cwd.split(os.sep)[-1]
        setting_module = '.'.join([cwd_module, 'settings'])
        try:
            sys.path.append(cwd[:cwd.rindex(os.sep)]+os.sep)
            _m = load_py(setting_module)
            if getattr(_m, 'PROJECT_NAME'):
                os.environ[APSMP] = setting_module
                if cwd not in sys.path and setPath:
                    sys.path.append(cwd)
                return True
        except:
            return False
    return False

def _get_project_settings():
    if APSMP not in os.environ:
        if not _init_project():
            return None
    path = os.environ.get(APSMP)
    _settings = Settings()
    _settings.set_module(path)
    return _settings

def _print_header(settings, inproject):
    if inproject:
        print("Amipy - project: %s\n" % (settings['project'].PROJECT_NAME))
    else:
        print("Amipy - no active project\n")

def _print_commands(settings, inproject,cmds):
    _print_header(settings, inproject)
    print("Usage:")
    print("  amipy <command> [options] [args]\n")
    print("Available commands:")
    for cmdname, cmd in sorted(cmds.items()):
        print("  %-13s %s" % (cmdname, cmd.short_desc()))
    if not inproject:
        print("\n  [ more ]    More commands available "
              "when run from project directory")
    print('\nUse "amipy <command> -h" to see more info about a command')

def _print_unknown_command(settings, cmdname, inproject):
    _print_header(settings, inproject)
    print("Unknown command: %s\n" % cmdname)
    print('Use "amipy" to see available commands')

def in_project():
    try:
        path = os.environ.get(APSMP)
        load_py(path)
        return True
    except:
        return False

def run(args=None,settings=None):
    if args is None:
        args = sys.argv
    if settings is None or not isinstance(settings,Settings):
        settings = _get_project_settings()
    install_logger(settings)
    inproject = in_project()
    cmds = _get_all_commands(settings,inproject)
    cmdname = _pop_command_name(args)
    if not cmdname:
        _print_commands(settings, inproject,cmds)
        sys.exit(0)
    elif cmdname not in cmds:
        _print_unknown_command(settings, cmdname, inproject)
        sys.exit(1)
    cmdcls = cmds[cmdname]
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(),
        conflict_handler='resolve')
    cmd =cmdcls(parser,cmdname)
    parser.usage = 'amipy %s %s '%(cmdname,cmd.syntax())
    parser.description = cmd.long_desc()
    cmd.add_options(parser)
    opts, argvs = parser.parse_args(args=args[1:])
    _execute(cmd,settings,opts, argvs)

def _execute(cmd,settings,opts, argvs):
    from amipy.exceptions import CommandUsageError
    try:
        cmd.handle(settings, opts, argvs)
    except CommandUsageError as e:
        print('Tip:wrong usage of command "%s".Correct usage e.g.\n'%e.cmd.name)
        e.parser.print_help()





from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):


    def parse(self,cmdnae,args,spiders):
        prompt = ''
        for _,i in enumerate(spiders):
            p = f'''
            \r\n*{'-'*60}*
            \r\n{_+1}. {i.name} | {i.status} | {i.__class__.__name__} | Success:{i._success}  Fail:{i._fail}  Exception:{i._exc}
            '''
            prompt += p
        return prompt





#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-16
'''
import amipy
import traceback
import warnings
from collections import defaultdict
from amipy.util.load import walk_modules,iter_spider_classes,load_py
from amipy.log import getLogger

class SpiderLoader(object):

    def __init__(self,settings):
        self.settings = settings
        self._spiders = {}
        self.project_path = settings['project'].PROJECT_NAME
        self._found=defaultdict(list)
        self.warn_only = True
        self.logger = getLogger(__name__)

    @property
    def spiders(self):
        return self._spiders.values()

    @spiders.setter
    def spiders(self,spiders):
        if not isinstance(spiders,dict):
            raise TypeError('Not a valid dict list of spiders.')
        for name,cls in spiders.items():
            if not isinstance(name,str):
                raise TypeError('Not a valid spider name type,got "%s".'
                                %type(name).__name__)
            if not issubclass(cls, amipy.Spider):
                raise TypeError('Not a valid Spider subclass,got "%s".'
                                %type(cls).__name__)
            if name in self._spiders:
                import warnings
                warnings.warn('Spider "%s" already exists.'%name)
                continue
            self._spiders.update({name:cls})

    def _singleton_spider(self,cls_):
        def _new(cls,*args,**kwargs):
            if not hasattr(cls, '_instance'):
                cls._instance = super(cls_, cls).__new__(cls)
            return cls._instance
        return _new

    def _load_spider_settings(self,path):
        try:
            settings = load_py(path)
        except ImportError as e:
            import warnings
            warnings.warn(
                'No settings module "%s" found,amipy will take "%s" instead.'
                %(path,self.settings['project'].SPIDER_BACKUP_SETTINGS))
            settings = load_py(self.settings['project'].SPIDER_BACKUP_SETTINGS)
        return settings

    def _load_spiders(self, module):
        for spcls in iter_spider_classes(module):
            spcls.__new__ = self._singleton_spider(spcls)
            self._found[spcls.name].append((module.__name__, spcls.__name__))
            spcls.__path = module.__name__[:module.__name__.rindex('.')]
            spcls.settings = self._load_spider_settings(
                '.'.join([spcls.__path,'settings']))
            self._spiders[spcls.name] = spcls
            self.logger.debug(f'Loaded spider "{spcls.__name__}:{spcls.name}"')

    def load_all_spiders(self):
        try:
            for module in walk_modules(self.project_path):
                self._load_spiders(module)
        except ImportError as e:
            if self.warn_only:
                msg = ("\n{tb}Could not load spiders from module '{modname}'. "
                       "See above traceback for details.".format(
                    modname=self.project_path, tb=traceback.format_exc()))
                warnings.warn(msg, RuntimeWarning)
            else:
                raise
        else:
            return [i() for i in self.spiders]



from distutils.core import  setup
from setuptools import find_packages

setup(
    name='Amipy',
    version='1.0.2',
    url='https://github.com/01ly/Amipy',
    description='A micro asynchronous Python website crawler framework',
    # long_description=open('README.md').read(),
    author='linkin',
    maintainer='linkin',
    maintainer_email='yooleak@outlook.com ',
    license='MIT',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': ['amipy = amipy.cmd:run']
    },
    install_requires=[
            'lxml',
            'bs4',
            'six>=1.5.2',
            'bloompy>=0.1.1',
        ],
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.5',
)


#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-15
'''
import amipy
from amipy.BaseClass import Hub
from amipy.middlewares import MiddleWareManager
from amipy.util.load import load_py
from amipy.log import getLogger

class SpiderHub(Hub):

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls,'_instance'):
            cls._instance = super(SpiderHub, cls).__new__(cls)
        return cls._instance

    def __init__(self,settings,crawler):
        super(SpiderHub, self).__init__()
        self.settings = settings
        self._success_counter = 0
        self._failed_counter = 0
        self._exception_counter = 0
        self.active = False
        self.looper = None
        self._crawler = crawler
        self.logger = getLogger(__name__)
        self._set_queue()

    def _set_queue(self):
        _queue = self.settings.gets('PROJECT_REQUESTS_QUEUE')
        self.requests = load_py(_queue)()
        self.logger.debug(f'Loaded Requests Queue:{type(self.requests).__name__}')

    def start(self,looper):
        self.looper = looper
        self.active = True
        for i in self.spiders:
            for seed in i.start_requests():
                i.status = 'RUNNING'
                if isinstance(seed, amipy.Request):
                    self.requests.put_nowait(seed)
        if self.requests.empty():
            print(f'* No start requests.Shutting down Amipy.\r\n')
            raise StopAsyncIteration
        self.logger.info(f'Got {self.requests.qsize()} start requests.')

    def takeover(self,spiders):
        self.spiders =spiders
        self.logger.debug(f'Takeover:{[i.name+":"+i.__class__.__name__ for i in spiders]}')
        self._binding()

    def _binding(self):
        for spider in self.spiders:
            spider.binding_hub = self
            spider.status = 'BOUND'
            self.priorities += spider.priority

    def accept(self,request):
        _all_req = []
        if isinstance(request,list):
            for req in request:
                if not isinstance(req, amipy.Request):
                   continue
                else:
                    _all_req.append(req)
        elif isinstance(request, amipy.Request):
            _all_req.append(request)
        return _all_req

    @MiddleWareManager.handle_resp
    def delegate(self,response):
        _res = []
        req = response.request
        spider = response.spider
        if response.status == 200:
            self._success_counter += 1
            spider._success += 1
            self.logger.info(f'[Success]{spider.name} {req.method}-{req.url}')
            a = self.accept(response.callback(response))
        elif response.status == -1:
            self._exception_counter += 1
            spider._exc +=1
            self.logger.info(f'[{response.exception.__class__.__name__}] {spider.name}'
                             f' {req.method}-{req.url} ')
            a = self.accept(response.excback(response))
        else:
            self._failed_counter += 1
            spider._fail += 1
            self.logger.info(f'[{response.status} Error]{spider.name} {req.method}-{req.url}')
            a = self.accept(response.errback(response))
        _res.extend(a)
        [self.requests.put_nowait(i) for i in _res if i]

    def __str__(self):
        return f'<SpiderHub obj at {hex(id(self))} active:{self.active}' \
               f' [spiders:{len(self.spiders)} success:{self._success_counter} ' \
               f'fail:{self._failed_counter} exc:{self._exception_counter}]>'




from amipy.middlewares import Middleware
from amipy.exceptions import DropRequest
from amipy import Url

class RulesHandleMiddleware(Middleware):

    def process_request(self,request):
        spider = request.spider
        url = request.url
        rules = [i for i in spider.rules if isinstance(i,Url)]
        for U in rules:
            if U.match(url):
                if U.drop:
                    raise DropRequest
                if U.cookies:
                    request._load_cookies(U.cookies)
                if U.cb:
                    cb = getattr(spider,U.cb,None)
                    if callable(cb):
                        request.callback = cb
                if bool(U.obey_robots_txt):
                    request.obey_robots_txt = True
                elif U.obey_robots_txt == False:
                    request.obey_robots_txt = False
                if U.down_type != None:
                    request.down_type = U.down_type
                if U.proxy != None:
                    request.proxy = U.proxy
                if U.proxy_auth != None:
                    request.proxy_auth = U.proxy_auth
                break
            else:
                if U.unmatch is None:
                    continue
                cb = getattr(spider,U.unmatch,None)
                if callable(cb):
                    request.callback = cb
        return request


import amipy

def send(request_or_list):
    if isinstance(request_or_list,list):
        for request in request_or_list:
            if not isinstance(request, amipy.Request):
                print(f'[TypeError] not a valid Request to send,'
                                f'got "{type(request).__name__}".')
                continue
            spider = request.spider
            if not isinstance(spider.binding_hub, amipy.SpiderHub):
                print('[TypeError] Not a valid binging SpiderHub for Spider %s,got "%s".'
                                % (spider.name, type(spider.binding_hub).__name__))
                continue
            spider.binding_hub.requests.put_nowait(request)
    elif isinstance(request_or_list, amipy.Request):
        spider = request_or_list.spider
        spider.binding_hub.requests.put_nowait(request_or_list)


#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-16
'''

from amipy.util.load import load_py

class Settings(dict):

    def __init__(self):
        super(Settings,self).__init__()
        self['project'] = type('project_settings',(),{})

    def set_module(self,path,level='project'):
        _module = load_py(path)
        self[level] = _module

    def gets(self,name,level='project'):
        try:
            return getattr(self[level],name)
        except Exception as e:
            raise AttributeError('No attribute "%s" in settings "%s" '%(name,self[level]))








import re

def gen_proxy(proxy,kind='text/html'):
    auth = extract_auth(proxy)
    _proxy = extract_ip_port(proxy)
    proxy = _proxy.strip('\n')
    if kind == 'text/html':
        return f'http://{auth}{proxy}'
    if kind == 'media':
        return {
            'http'  : f'http://{auth}{proxy}',
            'https' : f'https://{auth}{proxy}',
        }

def  is_proxy_valid(proxy):
    return re.findall(r'\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b\:\d+',proxy)

def extract_auth(proxy):
    res =  re.findall(r'//(.+)@',proxy)
    if res:
        return res[0]+'@'
    return ''

def extract_ip_port(proxy):
    if isinstance(proxy,dict):
        http_proxy = proxy['http']
        return is_proxy_valid(http_proxy)[0]
    else:
        http_proxy = is_proxy_valid(proxy)
        if http_proxy:
            return http_proxy[0]
        else:
            return proxy



#coding:utf-8
import os
import pickle
import amipy
import w3lib.url as urltool

class Request(object):

    def __init__(self,spider,url,*,
                 callback=None,
                 headers=None,
                 errback=None,
                 excback=None,
                 params=None,
                 proxy=None,
                 proxy_auth=None,
                 kwargs_cb=None,
                 cookies=None,
                 data=None,
                 down_type=None,
                 filter=None,
                 fingerprint=None,
                 timeout=None,
                 obey_robots_txt=None,
                 ignore = False,
                 save_path=None,
                 encoding='utf-8',
                 method='GET',
                 retry=0,
                 priority=0,
                 delay = 0,):

        assert isinstance(spider, amipy.Spider),\
            f'the Request should be bounded to a Spider,' \
            f'got "{type(spider).__name__}".'
        assert isinstance(kwargs_cb,dict) or kwargs_cb is  None,\
            'params from Request to success callback should be a dict.'
        assert isinstance(proxy_auth, dict) or proxy_auth is None, \
            'Auth for proxy should be a dict.'

        callback = callback if callback else spider.parse
        errback = errback if errback else spider.error
        excback = excback if excback else spider.exception

        assert callback or kwargs_cb is not None,\
            'there is no callback function for the Request.'
        assert callable(callback),\
            f'callback should be a coroutine function ,' \
            f'got "{type(callback).__name__}".'
        assert callable(errback) or errback is None, \
            f'errback should be a coroutine function,' \
            f'got "{type(errback).__name__}".'
        assert callable(excback) or excback is None, \
            f'excback should be a coroutine function,' \
            f'got "{type(excback).__name__}" .'
        assert method.strip().upper() in \
               ('GET','POST','HEAD','PUT','DELETE','UPDATE'), \
            "the method of a Request should be one of " \
            "the ['GET','POST','HEAD','PUT','DELETE','UPDATE']. "

        self.spider = spider
        self.callback = callback
        self.errback = errback
        self.excback = excback
        self.encoding = encoding
        self.priority = priority
        self.method = method
        self.data = data
        self.params = params
        self.fingerprint = fingerprint
        self.headers = headers
        self.proxy = proxy
        self.proxy_auth = proxy_auth
        self._ignore = ignore
        self.save_path = save_path
        self._tried = 0
        self.session = spider.session

        self.filter  = bool(filter) if filter != None \
            else spider.settings.BLOOMFILTER_URL_ON
        self.obey_robots_txt = bool(obey_robots_txt) if \
            obey_robots_txt != None else spider.settings.ROBOTS_TXT_OBEY
        self.kwargs_cb = {} if not kwargs_cb \
            else kwargs_cb
        self.down_type = down_type if down_type != None\
            else spider.settings.DEFAULT_DOWNLOAD_TYPE
        self.retry = retry if retry !=0 \
            else spider.settings.REQUEST_RETRY
        self.delay = delay if delay !=0 \
            else spider.settings.REQUEST_DELAY
        self.timeout = timeout if timeout is not None \
            else spider.settings.REQUEST_TIMEOUT

        self._set_url(url)
        self._load_cookies(cookies)
        self._set_delegate(spider.binding_hub.delegate)

    def _set_url(self,url):
        self.url = urltool.canonicalize_url(
            urltool.safe_download_url(url),encoding=self.encoding)

    def _load_cookies(self,cookies):
        if isinstance(cookies,dict):
            self.session.cookie_jar.update_cookies(cookies)
        elif isinstance(cookies,str):
            if os.path.exists(cookies):
                with open(cookies,'rb') as f:
                    rawdata = pickle.load(f)
                try:
                    self.session.cookie_jar.update_cookies(rawdata)
                except:
                    try:
                        _c_cookie = {j.key:j.value for i in rawdata.values()
                                     for j in i.values()}
                        self.session.cookie_jar.update_cookies(_c_cookie)
                    except:
                        return

    def _set_delegate(self,func):
        self.delegate_func = func

    @property
    def delegate_func(self):
        return self._func

    @delegate_func.setter
    def delegate_func(self,func):
        if func is None:
            self._func = None
        elif not callable(func):
            raise TypeError('delegate func should be callable,got "%s".'\
                            %type(func).__name__)
        else:
            self._func = func

    @property
    def url(self):
        return self.__url

    @url.setter
    def url(self,url):
        #only starts with schema:file,http,https allowed to be a valid url
        if not urltool.is_url(url):
            raise ValueError('Not a valid url for Request.Got:%s'%url)
        else:
            self.__url = urltool.safe_download_url(url)

    def __str__(self):
        return '<Request obj at %s [ spider=%s url=%s ] >'\
               %(hex(id(self)),self.spider.name,self.url)

    def __gt__(self, other):
        return self.spider > other.spider or \
               self.priority < other.priority




import threading, time
from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):

    def parse(self,cmdname,args,spiders):
        argv = args[0]
        if argv == 'spiders':
            return '\r\n Not a valid usage of command: resume <spider name>'
        d = {i.name:i for i in spiders}
        if argv in d:
            spider = d[argv]
            if spider.status == 'RUNNING':
                return f"* Spider {argv} is running,it can not resume. "
            elif spider.status == 'PAUSE':
                lock = threading.Lock()
                lock.acquire()
                spider.status = 'RESUME'
                spider._resume_at = time.ctime()
                lock.release()
                return f'* Spider {argv} resumed at {spider._resume_at} successfully.'
            elif spider.status == 'RESUME':
                return f'* Spider {argv} is resuming.'
            elif spider.status == 'CLOSE':
                return f'* Spider {argv} is closed.'
            elif spider.status == 'STOP':
                return f'* Spider {argv} is stopped.Using "restart" command to restart it.'
            else:
                return f'* Invalid resuming status "{spider.status}" for a spider.'
        else:
            return f'* No spider "{argv}" in the project.'




from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):

    def parse(self,cmdname,args,spiders):
        argv = args[0]
        if argv == 'spiders':
            return '\r\n Not a valid usage of command: echo <spider name>'
        prompt = ''
        d = {i.name:i for i in spiders}
        if argv in d:
            spider = d[argv]
            cd = f'''
            \r\n- UrlFilter:count[{spider.urlfilter.count}] capacity[{spider.urlfilter.capacity}] error_rate[{spider.urlfilter.error_rate}]
            \r\n- HtmlFilter:count[{spider.respfilter.count}] capacity[{spider.respfilter.capacity}] error_rate[{spider.respfilter.error_rate}]
            ''' if spider.urlfilter and spider.respfilter else ''
            sp = f'''
            \r\n----------------Spider-{argv}-------------------
            \r\n- Name:{spider.name}  Status:{spider.status}
            \r\n- Class:{spider.__class__.__name__}
            \r\n- Success:{spider._success}    Fail:{spider._fail}     Exception:{spider._exc}'''+ cd +f'''
            \r\n- Priority:{spider.priority} 
            \r\n- SeedUrls:{spider.urls}
            \r\n- Path:{spider.settings.PATH}
            \r\n- Session:{spider.session}
            \r\n- StartAt:{spider._start_at}
            \r\n- PausedAt:{spider._pause_at}
            \r\n- ResumeAt:{spider._resume_at}
            \r\n- StopAt:{spider._stop_at}
            \r\n- RestartAt:{spider._restart_at}
            \r\n- CloseAt:{spider._close_at}
            \r\n{'-'*50}
            '''
            prompt += sp
        else:
            return f'* No spider "{argv}" in the project.'
        return prompt




import threading, time
from amipy.BaseClass import SpiderClientCommand

class SCommand(SpiderClientCommand):

    def parse(self,cmdname,args,spiders):
        argv = args[0]
        if argv == 'spiders':
            return '\r\n Not a valid usage of command: close <spider name>'
        d = {i.name:i for i in spiders}
        if argv in d:
            spider = d[argv]
            if spider.status == 'CLOSE':
                return f"* Spider {argv} hab been closed at {spider._close_at} already. "
            else:
                lock = threading.Lock()
                lock.acquire()
                spider.status = 'CLOSE'
                spider._close_at = time.ctime()
                lock.release()
                return f'* {spider._close_at} Spider {argv} closed successfully. '
        else:
            return f'* No spider "{argv}" in the project.'



import hashlib
from bloompy import SCBloomFilter

def _generate_filter(settings):
    error_rate = settings.BLOOMFILTER_ERROR_RATE
    capacity = settings.BLOOMFILTER_INITIAL_CAPACITY
    return SCBloomFilter(error_rate,initial_capacity=capacity,mode=1)

def _to_feature(request):
    url =request.url
    method = request.method
    params = '' if not request.params else request.params
    data = '' if not request.data else request.data
    _str = ''.join([method, url, str(params), str(data)])
    _feature = _to_md5(_str)
    return _feature

def _to_md5(_str,encoding='utf-8'):
    bytes_like = bytes(_str, encoding=encoding) if \
        isinstance(_str,str) else _str
    b_md5 = hashlib.md5(bytes_like).hexdigest()
    return b_md5


import sys
import logging

__all__ = ['getLogger','install_logger']

def getLogger(name):
    logger = Logger(name)
    return logger.logger

def install_logger(settings):
    Logger.settings = settings

class Logger(object):

    settings = None
    encoding = 'utf-8'

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super(Logger, cls).__new__(cls)
        return cls._instance

    def __init__(self,name):
        self.logger = logging.getLogger(name)
        self._set_logger(self.logger)

    def _set_logger(self,logger):
        logger.setLevel(logging.DEBUG)
        if logger.hasHandlers():
            return
        for i in self._get_handlers():
            logger.addHandler(i)

    @classmethod
    def _get_handlers(cls):
        enabled = cls.settings.gets('LOG_ENABLE')
        handlers = []
        if enabled:
            _level = cls.settings.gets('LOG_LEVEL')
            f_path = cls.settings.gets('LOG_FILE_SAVE_PATH')
            encoding = cls.settings.get('LOG_FILE_ENCODING')
            _formatter = cls.settings.gets('LOG_FORMAT')[_level]
            _datefmt = cls.settings.gets('LOG_DATE_FORMAT')
            level = getattr(logging,_level)
            formatter = logging.Formatter(_formatter,datefmt=_datefmt)
            if f_path:
                encoding = encoding if encoding else cls.encoding
                f_handler = logging.FileHandler(f_path,
                                                encoding=encoding)
                f_handler.setLevel(logging.DEBUG)
                f_handler.setFormatter(formatter)
                handlers.append(f_handler)
            s_handler = logging.StreamHandler(stream=sys.stdout)
            s_handler.setLevel(level)
            s_handler.setFormatter(formatter)
            handlers.append(s_handler)
        else:
            handlers.append(logging.NullHandler())
        return handlers






#coding:utf-8
'''
    author : linkin
    e-mail : yooleak@outlook.com
    date   : 2018-11-17
'''

import asyncio
from inspect import iscoroutine
from amipy.log import getLogger

class Looper(object):
    def __init__(self):
        self.loop = asyncio.get_event_loop()
        self.logger = getLogger(__name__)

    def run_coroutine(self,coroutines):
        if isinstance(coroutines,list):
            tasks = [asyncio.ensure_future(i) for i in coroutines if iscoroutine(i)]
            _coroutine = asyncio.gather(*tasks)
        elif iscoroutine(coroutines):
            _coroutine = coroutines
        else:
            raise TypeError('Not a coroutine or coroutine list to run with,got "%s".'%type(coroutines).__name__)
        try:
            results = self.loop.run_until_complete(_coroutine)
        except Exception as e:
            _coroutine.cancel()
        else:
            return results

    def run_forever(self):
        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            self.stop()

    def run_tasks(self,tasks):
        if tasks is None:
            return
        try:
            res = self.loop.run_until_complete(asyncio.gather(*tasks))
            return res
        except KeyboardInterrupt:
            self.logger.info('Shutting  down Amipy.')
            self.stop()
            raise StopAsyncIteration

    def stop(self):
        tasks = asyncio.Task.all_tasks(loop=self.loop)
        group = asyncio.gather(*tasks, return_exceptions=True)
        group.cancel()
        self.loop.stop()
        self.logger.debug(f'Gathered {len(tasks)} Tasks.')
        self.logger.debug(f'Task Group {group} has been canceled')
        self.logger.debug('Stopped looper.')






